import json
import pathlib

import pytest

from manubot.cite.citeproc import remove_jsonschema_errors

directory = pathlib.Path(__file__).parent
csl_instances = [
    x.name for x in directory.glob('csl-json/*-csl')
]


@pytest.mark.parametrize('name', csl_instances)
def test_remove_jsonschema_errors(name):
    """
    Tests whether CSL pruning of raw.json produces the expected pruned.json.
    The fidelity of the remove_jsonschema_errors implementation is
    difficult to assess theoretically, therefore its performance should be
    evaluated empirically with thorough test coverage. It is recommended that
    all conceivable conformations of invalid CSL that citeproc methods may
    generate be tested.

    To create a new test case, derive pruned.json from raw.json, by manually
    deleting any invalid fields. Do not use `manubot cite` to directly generate
    pruned.json as that also relies on remove_jsonschema_errors for pruning.
    """
    data_dir = directory / 'csl-json' / name
    raw = json.loads(data_dir.joinpath('raw.json').read_text())
    expected = json.loads(data_dir.joinpath('pruned.json').read_text())
    pruned = remove_jsonschema_errors(raw)
    assert pruned == expected



import json
import logging
import urllib.request

import requests

from manubot.cite.pubmed import get_pubmed_ids_for_doi
from manubot.util import get_manubot_user_agent


def expand_short_doi(short_doi):
    """
    Convert a shortDOI to a regular DOI.
    """
    if not short_doi.startswith('10/'):
        raise ValueError(f'shortDOIs start with `10/`, but expand_short_doi received: {short_doi}')
    url = f'https://doi.org/api/handles/{short_doi.lower()}'
    params = {
        "type": "HS_ALIAS",
    }
    response = requests.get(url, params=params)
    # response documentation at https://www.handle.net/proxy_servlet.html
    results = response.json()
    response_code = results.get('responseCode')  # Handle protocol response code
    if response_code == 100:
        raise ValueError(f'Handle not found. Double check short_doi: {short_doi}')
    if response_code == 200:
        raise ValueError(f'HS_ALIAS values not found. Double check short_doi: {short_doi}')
    if response_code != 1:
        raise ValueError(f'Error response code of {response_code} returned by {response.url}')
    values = results.get('values', [])
    for value in values:
        if value.get('type') == 'HS_ALIAS':
            doi = value['data']['value']
            return doi.lower()
    raise RuntimeError(
        f'HS_ALIAS value not found by expand_short_doi("{short_doi}")\n'
        f'The following JSON was retrieved from {response.url}:\n'
        + json.dumps(results, indent=2)
    )


def get_short_doi_url(doi):
    """
    Get the shortDOI URL for a DOI.
    """
    quoted_doi = urllib.request.quote(doi)
    url = 'http://shortdoi.org/{}?format=json'.format(quoted_doi)
    headers = {
        'User-Agent': get_manubot_user_agent(),
    }
    try:
        response = requests.get(url, headers=headers).json()
        short_doi = response['ShortDOI']
        short_url = 'https://doi.org/' + short_doi[3:]  # Remove "10/" prefix
        return short_url
    except Exception:
        logging.warning(f'shortDOI lookup failed for {doi}', exc_info=True)
        return None


def get_doi_citeproc(doi):
    """
    Use Content Negotioation (http://citation.crosscite.org/docs.html) to
    retrieve the citeproc JSON citation for a DOI.
    """
    url = 'https://doi.org/' + urllib.request.quote(doi)
    header = {
        'Accept': 'application/vnd.citationstyles.csl+json',
        'User-Agent': get_manubot_user_agent(),
    }
    response = requests.get(url, headers=header)
    try:
        citeproc = response.json()
    except Exception as error:
        logging.error(f'Error fetching metadata for doi:{doi}.\n'
                      f'Invalid response from {response.url}:\n{response.text}')
        raise error
    citeproc['URL'] = f'https://doi.org/{doi}'
    short_doi_url = get_short_doi_url(doi)
    if short_doi_url:
        citeproc['URL'] = short_doi_url
    try:
        citeproc.update(get_pubmed_ids_for_doi(doi))
    except Exception:
        logging.warning(f'Error calling get_pubmed_ids_for_doi for {doi}', exc_info=True)
    return citeproc



import logging


def cli_process(args):
    args_dict = vars(args)

    # Set paths for content
    content_dir = args.content_directory
    if not content_dir.is_dir():
        logging.warning(f'content directory does not exist: {content_dir}')
    args_dict['citation_tags_path'] = content_dir.joinpath('citation-tags.tsv')
    args_dict['meta_yaml_path'] = content_dir.joinpath('metadata.yaml')
    args_dict['manual_references_path'] = content_dir.joinpath('manual-references.json')

    # Set paths for output
    output_dir = args.output_directory
    output_dir.mkdir(exist_ok=True)
    args_dict['manuscript_path'] = output_dir.joinpath('manuscript.md')
    args_dict['citations_path'] = output_dir.joinpath('citations.tsv')
    args_dict['references_path'] = output_dir.joinpath('references.json')
    args_dict['variables_path'] = output_dir.joinpath('variables.json')

    # Set paths for caching
    args_dict['cache_directory'] = args.cache_directory or output_dir
    args.cache_directory.mkdir(exist_ok=True)
    args_dict['requests_cache_path'] = str(args.cache_directory.joinpath('requests-cache'))

    from manubot.process.util import prepare_manuscript
    prepare_manuscript(args)



import pytest

from manubot.cite.isbn import (
    get_isbn_citeproc_citoid,
    get_isbn_citeproc_isbnlib,
    get_isbn_citeproc_zotero,
)


@pytest.mark.xfail(reason="Quotation in title removed at some upstream point")
def test_citation_to_citeproc_isbnlib_title_with_quotation_mark():
    csl_item = get_isbn_citeproc_isbnlib('9780312353780')
    assert csl_item['type'] == 'book'
    assert csl_item['title'].startswith('"F" is for Fugitive')


def test_get_isbn_citeproc_citoid_weird_date():
    """
    isbn:9780719561023 has a date value of "(2004 printing)"
    https://en.wikipedia.org/api/rest_v1/data/citation/mediawiki/9780719561023
    """
    csl_item = get_isbn_citeproc_citoid('9780719561023')
    assert csl_item['issued']['date-parts'] == [[2004]]
    assert csl_item['ISBN'] == '9780719561023'


def test_get_isbn_citeproc_citoid_not_found():
    """
    isbn:9781439566039 is not found by Citoid:
    https://en.wikipedia.org/api/rest_v1/data/citation/mediawiki/9781439566039
    """
    with pytest.raises(KeyError, match=r'Metadata for ISBN [0-9]{10,13} not found'):
        get_isbn_citeproc_citoid('9781439566039')


def test_get_isbn_citeproc_zotero_with_note_issue():
    """
    translation-server returns two metadata records for this ISBN.
    The second has itemType=note and previousely caused CSL export to fail.
    https://github.com/zotero/translation-server/issues/67
    """
    isbn = '9780262517638'
    csl_item = get_isbn_citeproc_zotero(isbn)
    assert csl_item['author'][0]['family'] == 'Suber'



from manubot.process.manuscript import (
    get_citation_strings,
    replace_citations_strings_with_ids
)


def test_get_citation_strings_1():
    text = '''
    Sci-Hub has released article request records from its server logs, covering 165 days from September 2015 through February 2016 [@doi:10.1126/science.352.6285.508; @doi:10.1126/science.aaf5664; @doi:10.5061/dryad.q447c/1].
    We filtered for valid requests by excluding DOIs not included in our literature catalog and omitting requests that occurred before an article's publication date.

    Figure {@fig:citescore}B shows that articles from highly cited journals were visited much more frequently on average.

    @10.5061/bad_doi says blah but @url:https://www.courtlistener.com/docket/4355308/1/elsevier-inc-v-sci-hub/ disagrees.
    '''
    citations = get_citation_strings(text)
    expected = sorted([
        '@doi:10.1126/science.352.6285.508',
        '@doi:10.1126/science.aaf5664',
        '@doi:10.5061/dryad.q447c/1',
        '@url:https://www.courtlistener.com/docket/4355308/1/elsevier-inc-v-sci-hub/',
    ])
    assert citations == expected


def test_replace_citations_strings_with_ids():
    """
    Test that text does not get converted to:

    > our new Manubot tool [@cTN2TQIL-rootstock; @cTN2TQIL] for automating
    manuscript generation.

    See https://github.com/manubot/manubot/issues/9
    """
    string_to_id = {
        '@url:https://github.com/manubot/manubot': 'mNMayr3f',
        '@url:https://github.com/greenelab/manubot-rootstock': '1B7Y2HVtw',
    }
    text = 'our new Manubot tool [@url:https://github.com/greenelab/manubot-rootstock; @url:https://github.com/manubot/manubot] for automating manuscript generation.'
    text = replace_citations_strings_with_ids(text, string_to_id)
    assert '[@1B7Y2HVtw; @mNMayr3f]' in text



import subprocess

import manubot


def test_version():
    stdout = subprocess.check_output(
        ['manubot', '--version'],
        universal_newlines=True,
    )
    version_str = f'v{manubot.__version__}'
    assert version_str == stdout.rstrip()


def test_missing_subcommand():
    process = subprocess.run(
        ['manubot'],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
    )
    print(process.stderr)
    assert process.returncode == 2
    assert 'error: the following arguments are required: subcommand' in process.stderr



import collections
import logging
import re
import xml.etree.ElementTree

import requests

from manubot.util import get_manubot_user_agent


def get_arxiv_citeproc(arxiv_id):
    """
    Return citeproc item for an arXiv record.

    arxiv_id can be versioned, like `1512.00567v2`, or versionless, like
    `1512.00567`. If versionless, the arXiv API will return metadata for the
    latest version. Legacy IDs, such as `cond-mat/0703470v2`, are also
    supported.

    If arXiv has an associated DOI for the record, a warning is logged to
    alert the user that an alternative version of record exists.

    References:
    https://arxiv.org/help/api/index
    http://citeproc-js.readthedocs.io/en/latest/csl-json/markup.html
    https://github.com/citation-style-language/schema/blob/master/csl-data.json
    """
    url = 'https://export.arxiv.org/api/query'
    params = {
        'id_list': arxiv_id,
        'max_results': 1,
    }
    headers = {
        'User-Agent': get_manubot_user_agent(),
    }
    response = requests.get(url, params, headers=headers)

    # XML namespace prefixes
    prefix = '{http://www.w3.org/2005/Atom}'
    alt_prefix = '{http://arxiv.org/schemas/atom}'

    # Parse XML
    xml_tree = xml.etree.ElementTree.fromstring(response.text)
    entry, = xml_tree.findall(prefix + 'entry')

    # Create dictionary for CSL Item
    csl_item = collections.OrderedDict()

    # Extract versioned arXiv ID
    url = entry.findtext(prefix + 'id')
    pattern = re.compile(r'arxiv.org/abs/(.+)')
    match = pattern.search(url)
    versioned_id = match.group(1)
    csl_item['number'] = versioned_id
    _, csl_item['version'] = versioned_id.rsplit('v', 1)
    csl_item['URL'] = 'https://arxiv.org/abs/' + versioned_id

    # Extrat CSL title field
    csl_item['title'] = entry.findtext(prefix + 'title')

    # Extract CSL date field
    published = entry.findtext(prefix + 'published')
    published, _ = published.split('T', 1)
    csl_item['issued'] = {'date-parts': [[int(x) for x in published.split('-')]]}

    # Extract authors
    authors = list()
    for elem in entry.findall(prefix + 'author'):
        name = elem.findtext(prefix + 'name')
        author = {'literal': name}
        authors.append(author)
    csl_item['author'] = authors

    # Set publisher to arXiv
    csl_item['container-title'] = 'arXiv'
    csl_item['publisher'] = 'arXiv'

    # Extract abstract
    abstract = entry.findtext(prefix + 'summary').strip()
    if abstract:
        csl_item['abstract'] = abstract

    # Check if the article has been published with a DOI
    DOI = entry.findtext('{http://arxiv.org/schemas/atom}doi')
    if DOI:
        csl_item['DOI'] = DOI
        journal_ref = entry.findtext(alt_prefix + 'journal_ref')
        msg = f'arXiv article {arxiv_id} published at https://doi.org/{DOI}'
        if journal_ref:
            msg += f' — {journal_ref}'
        logging.warning(msg)
    # Set CSL type to report for preprint
    csl_item['type'] = 'report'
    return csl_item



import json
import pathlib
import shutil
import subprocess

import pytest

from manubot.cite.cite_command import (
    _get_pandoc_info,
)


def test_cite_command_empty():
    process = subprocess.run(
        ['manubot', 'cite'],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
    )
    print(process.stderr)
    assert process.returncode == 2
    assert 'the following arguments are required: citations' in process.stderr


def test_cite_command_stdout():
    process = subprocess.run(
        ['manubot', 'cite', 'arxiv:1806.05726v1'],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
    )
    print(process.stderr)
    assert process.returncode == 0
    csl, = json.loads(process.stdout)
    assert csl['URL'] == 'https://arxiv.org/abs/1806.05726v1'


def test_cite_command_file(tmpdir):
    path = pathlib.Path(tmpdir) / 'csl-items.json'
    process = subprocess.run(
        ['manubot', 'cite', '--output', str(path), 'arxiv:1806.05726v1'],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    print(process.stderr.decode())
    assert process.returncode == 0
    with path.open() as read_file:
        csl, = json.load(read_file)
    assert csl['URL'] == 'https://arxiv.org/abs/1806.05726v1'


@pytest.mark.parametrize(['args', 'expected'], [
    ([], 'references-plain.txt'),
    (['--format', 'plain'], 'references-plain.txt'),
    (['--format', 'markdown'], 'references-markdown.md'),
    (['--format', 'html'], 'references-html.html'),
    (['--format', 'jats'], 'references-jats.xml'),
], ids=['no-args', '--format=plain', '--format=markdown', '--format=html', '--format=jats'])
@pytest.mark.skipif(
    not shutil.which('pandoc'),
    reason='pandoc installation not found on system'
)
@pytest.mark.skipif(
    not shutil.which('pandoc-citeproc'),
    reason='pandoc-citeproc installation not found on system'
)
def test_cite_command_render_stdout(args, expected):
    """
    Test the stdout output of `manubot cite --render` with various formats.
    The output is sensitive to the version of Pandoc used, so rather than fail when
    the system's pandoc is outdated, the test is skipped. 
    """
    pandoc_version = _get_pandoc_info()['pandoc version']
    for output in 'markdown', 'html', 'jats':
        if output in args and pandoc_version < (2, 5):
            pytest.skip(f"Test {output} output assumes pandoc >= 2.5")
    if pandoc_version < (2, 0):
        pytest.skip("Test requires pandoc >= 2.0 to support --lua-filter and --csl=URL")
    expected = (
        pathlib.Path(__file__).parent
        .joinpath('cite-command-rendered', expected)
        .read_text()
    )
    args = [
        'manubot', 'cite', '--render',
        '--csl', 'https://github.com/greenelab/manubot-rootstock/raw/e83e51dcd89256403bb787c3d9a46e4ee8d04a9e/build/assets/style.csl',
        'arxiv:1806.05726v1', 'doi:10.7717/peerj.338', 'pmid:29618526',
    ] + args
    process = subprocess.run(
        args,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
    )
    print(' '.join(process.args))
    print(process.stdout)
    print(process.stderr)
    assert process.stdout == expected


def teardown_module(module):
    """
    Avoid too many requests to NCBI E-Utils in the test_pubmed.py,
    which is executed following this module. E-Utility requests are
    capped at 3 per second, which is usually controled by _get_eutils_rate_limiter,
    but this does not seem to work across test modules.
    """
    import time
    time.sleep(1)



from manubot.cite.wikidata import get_wikidata_citeproc


def test_get_wikidata_citeproc():
    """
    Test metadata extraction from https://www.wikidata.org/wiki/Q50051684
    """
    wikidata_id = 'Q50051684'
    csl_item = get_wikidata_citeproc(wikidata_id)
    assert 'Sci-Hub provides access to nearly all scholarly literature' in csl_item['title']
    assert csl_item['container-title'] == 'eLife'
    assert csl_item['DOI'] == '10.7554/elife.32822'


def test_get_wikidata_citeproc_author_ordering():
    """
    Test extraction of author ordering from https://www.wikidata.org/wiki/Q50051684.
    Wikidata uses a "series ordinal" qualifier that must be considered or else author
    ordering may be wrong.

    Author ordering was previously not properly set by the Wikidata translator
    https://github.com/zotero/translators/issues/1790
    """
    wikidata_id = 'Q50051684'
    csl_item = get_wikidata_citeproc(wikidata_id)
    family_names =[author['family'] for author in csl_item['author']]
    print(family_names)
    assert family_names == [
        'Himmelstein',
        'Romero',
        'Levernier',
        'Munro',
        'McLaughlin',
        'Greshake',  # actually should be Greshake Tzovaras
        'Greene',
    ]



import pathlib
import re
import shlex
import subprocess

import pytest

readme_path = pathlib.Path(__file__).parent.parent.parent / 'README.md'
readme = readme_path.read_text()

template = r'''
<!-- test codeblock contains output of `{command}` -->
```
{output}```
'''
pattern = template.format(
    command=r"(?P<command>.+?)",
    output=r"(?P<output>.+?)",
)
pattern = re.compile(pattern, re.DOTALL)
matches = list(pattern.finditer(readme))


@pytest.mark.parametrize(
    argnames=['command', 'expected'],
    argvalues=[match.groups() for match in matches],
    ids=[match.group('command') for match in matches],
)
def test_readme_codeblock_contains_output_from(command, expected):
    """
    If this test fails, ensure that codeblocks in README.md have the correct
    output. To enable this check for output in a codeblock, use the following
    construct:

    <!-- test codeblock contains output of `{command}` -->
    ```
    {expected}
    ```
    """
    output = _get_output_from(command)
    assert output == expected


def _get_output_from(command):
    return subprocess.check_output(shlex.split(command), universal_newlines=True)


def _match_to_repl(match):
    template_dict = match.groupdict()
    template_dict['output'] = _get_output_from(template_dict['command'])
    return template.format(**template_dict)


if __name__ == '__main__':
    """
    Run `python tests/test_readme.py` to populate README codeblocks with
    output from the specified commands.
    """
    repl_readme = pattern.sub(repl=_match_to_repl, string=readme)
    readme_path.write_text(repl_readme)



import collections
import datetime
import json
import logging
import pathlib
import re

from manubot.cite.util import (
    citation_pattern,
    is_valid_citation_string,
)


def get_citation_strings(text):
    """
    Extract the deduplicated list of citations in a text. Citations that are
    clearly invalid such as `doi:/453` are not returned.
    """
    citations_strings = set(citation_pattern.findall(text))
    citations_strings = filter(
        lambda x: is_valid_citation_string(x, allow_tag=True, allow_raw=True, allow_pandoc_xnos=True),
        citations_strings,
    )
    return sorted(citations_strings)


def get_text(directory):
    """
    Return a concatenated string of section texts from the specified directory.
    """
    section_dir = pathlib.Path(directory)
    paths = sorted(section_dir.glob('[0-9]*.md'))
    name_to_text = collections.OrderedDict()
    for path in paths:
        name_to_text[path.stem] = path.read_text()
    logging.info('Manuscript content parts:\n' + '\n'.join(name_to_text))
    return '\n\n'.join(name_to_text.values()) + '\n'


def replace_citations_strings_with_ids(text, string_to_id):
    """
    Convert citations to their IDs for pandoc.

    `text` is markdown source text

    `string_to_id` is a dictionary like:
    @10.7287/peerj.preprints.3100v1 → 11cb5HXoY
    """
    for old, new in string_to_id.items():
        text = re.sub(
            pattern=re.escape(old) + r'(?![\w:.#$%&\-+?<>~/]*[a-zA-Z0-9/])',
            repl='@' + new,
            string=text,
        )
    return text


def get_manuscript_stats(text, citation_df):
    """
    Compute manuscript statistics.
    """
    stats = collections.OrderedDict()

    # Number of distinct references by type
    ref_counts = (
        citation_df
        .standard_citation
        .drop_duplicates()
        .map(lambda x: x.split(':')[0])
        .pipe(collections.Counter)
    )
    ref_counts['total'] = sum(ref_counts.values())
    stats['reference_counts'] = ref_counts
    stats['word_count'] = len(text.split())
    logging.info(f"Generated manscript stats:\n{json.dumps(stats, indent=2)}")
    return stats


def datetime_now():
    """
    Return the current datetime, with timezone awareness
    https://stackoverflow.com/a/39079819/4651668
    """
    tzinfo = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo
    return datetime.datetime.now(tzinfo)



def get_wikidata_citeproc(identifier):
    """
    Get a CSL JSON item with the citation metadata for a Wikidata item.
    identifier should be a Wikidata item ID corresponding to a citeable
    work, such as Q50051684.
    """
    url = f'https://www.wikidata.org/wiki/{identifier}'
    from manubot.cite.url import get_url_citeproc_zotero
    csl_item = get_url_citeproc_zotero(url)
    if 'DOI' in csl_item:
        csl_item['DOI'] = csl_item['DOI'].lower()
    if not 'URL' in csl_item:
        csl_item['URL'] = url
    return csl_item



import pytest

from manubot.cite.util import (
    citation_pattern,
    citation_to_citeproc,
    get_citation_id,
    inspect_citation_identifier,
    standardize_citation,
)


@pytest.mark.parametrize("citation_string", [
    ('@doi:10.5061/dryad.q447c/1'),
    ('@arxiv:1407.3561v1'),
    ('@doi:10.1007/978-94-015-6859-3_4'),
    ('@tag:tag_with_underscores'),
    ('@tag:tag-with-hyphens'),
    ('@url:https://greenelab.github.io/manubot-rootstock/'),
    ('@tag:abc123'),
    ('@tag:123abc'),
])
def test_citation_pattern_match(citation_string):
    match = citation_pattern.fullmatch(citation_string)
    assert match


@pytest.mark.parametrize("citation_string", [
    ('doi:10.5061/dryad.q447c/1'),
    ('@tag:abc123-'),
    ('@tag:abc123_'),
    ('@-tag:abc123'),
    ('@_tag:abc123'),
])
def test_citation_pattern_no_match(citation_string):
    match = citation_pattern.fullmatch(citation_string)
    assert match is None


@pytest.mark.parametrize("standard_citation,expected", [
    ('doi:10.5061/dryad.q447c/1', 'kQFQ8EaO'),
    ('arxiv:1407.3561v1', '16kozZ9Ys'),
    ('pmid:24159271', '11sli93ov'),
    ('url:http://blog.dhimmel.com/irreproducible-timestamps/', 'QBWMEuxW'),
])
def test_get_citation_id(standard_citation, expected):
    citation_id = get_citation_id(standard_citation)
    assert citation_id == expected


@pytest.mark.parametrize("citation,expected", [
    ('doi:10.5061/DRYAD.q447c/1', 'doi:10.5061/dryad.q447c/1'),
    ('doi:10.5061/dryad.q447c/1', 'doi:10.5061/dryad.q447c/1'),
    ('doi:10/b6vnmd', 'doi:10.1016/s0933-3657(96)00367-3'),
    ('doi:10/B6VNMD', 'doi:10.1016/s0933-3657(96)00367-3'),
    ('doi:10/xxxxxxxxxxxxxYY', 'doi:10/xxxxxxxxxxxxxyy'),  # passthrough non-existent shortDOI
    ('pmid:24159271', 'pmid:24159271'),
    ('isbn:1339919885', 'isbn:9781339919881'),
    ('isbn:1-339-91988-5', 'isbn:9781339919881'),
    ('isbn:978-0-387-95069-3', 'isbn:9780387950693'),
    ('isbn:9780387950938', 'isbn:9780387950938'),
    ('isbn:1-55860-510-X', 'isbn:9781558605107'),
    ('isbn:1-55860-510-x', 'isbn:9781558605107'),
])
def test_standardize_citation(citation, expected):
    """
    Standardize identifiers based on their source
    """
    output = standardize_citation(citation)
    assert output == expected


@pytest.mark.parametrize('citation', [
    'doi:10.7717/peerj.705',
    'doi:10/b6vnmd',
    'pmcid:PMC4304851',
    'pmid:25648772',
    'arxiv:1407.3561',
    'isbn:978-1-339-91988-1',
    'isbn:1-339-91988-5',
    'wikidata:Q1',
    'wikidata:Q50051684',
    'url:https://peerj.com/articles/705/',
])
def test_inspect_citation_identifier_passes(citation):
    """
    These citations should pass inspection by inspect_citation_identifier.
    """
    assert inspect_citation_identifier(citation) is None


@pytest.mark.parametrize(['citation', 'contains'], [
    ('doi:10.771/peerj.705', 'Double check the DOI'),
    ('doi:10/b6v_nmd', 'Double check the shortDOI'),
    ('doi:7717/peerj.705', 'must start with `10.`'),
    ('doi:b6vnmd', 'must start with `10.`'),
    ('pmcid:25648772', 'must start with `PMC`'),
    ('pmid:PMC4304851', 'Should pmid:PMC4304851 switch the citation source to `pmcid`?'),
    ('isbn:1-339-91988-X', 'identifier violates the ISBN syntax'),
    ('wikidata:P212', 'item IDs must start with `Q`'),
    ('wikidata:QABCD', 'does not conform to the Wikidata regex'),
])
def test_inspect_citation_identifier_fails(citation, contains):
    """
    These citations should fail inspection by inspect_citation_identifier.
    """
    report = inspect_citation_identifier(citation)
    assert report is not None
    assert isinstance(report, str)
    assert contains in report


@pytest.mark.xfail(reason='https://twitter.com/dhimmel/status/950443969313419264')
def test_citation_to_citeproc_doi_datacite():
    citation = 'doi:10.7287/peerj.preprints.3100v1'
    citeproc = citation_to_citeproc(citation)
    assert citeproc['id'] == '11cb5HXoY'
    assert citeproc['URL'] == 'https://doi.org/10.7287/peerj.preprints.3100v1'
    assert citeproc['DOI'] == '10.7287/peerj.preprints.3100v1'
    assert citeproc['type'] == 'report'
    assert citeproc['title'] == 'Sci-Hub provides access to nearly all scholarly literature'
    authors = citeproc['author']
    assert authors[0]['family'] == 'Himmelstein'
    assert authors[-1]['family'] == 'Greene'


def test_citation_to_citeproc_arxiv():
    citation = 'arxiv:cond-mat/0703470v2'
    citeproc = citation_to_citeproc(citation)
    assert citeproc['id'] == 'ES92tcdg'
    assert citeproc['URL'] == 'https://arxiv.org/abs/cond-mat/0703470v2'
    assert citeproc['number'] == 'cond-mat/0703470v2'
    assert citeproc['version'] == '2'
    assert citeproc['type'] == 'report'
    assert citeproc['container-title'] == 'arXiv'
    assert citeproc['title'] == 'Portraits of Complex Networks'
    authors = citeproc['author']
    assert authors[0]['literal'] == 'J. P. Bagrow'
    assert citeproc['DOI'] == '10.1209/0295-5075/81/68004'


def test_citation_to_citeproc_pmc():
    """
    https://api.ncbi.nlm.nih.gov/lit/ctxp/v1/pmc/?format=csl&id=3041534
    """
    citation = f'pmcid:PMC3041534'
    citeproc = citation_to_citeproc(citation)
    assert citeproc['id'] == 'RoOhUFKU'
    assert citeproc['URL'] == 'https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3041534/'
    assert citeproc['container-title-short'] == 'Summit Transl Bioinform'
    assert citeproc['title'] == 'Secondary Use of EHR: Data Quality Issues and Informatics Opportunities'
    authors = citeproc['author']
    assert authors[0]['family'] == 'Botsis'
    assert citeproc['PMID'] == '21347133'
    assert citeproc['PMCID'] == 'PMC3041534'


def test_citation_to_citeproc_pubmed_1():
    """
    Generated from XML returned by
    https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=pubmed&id=21347133&rettype=full
    """
    citation = 'pmid:21347133'
    citeproc = citation_to_citeproc(citation)
    assert citeproc['id'] == 'y9ONtSZ9'
    assert citeproc['type'] == 'article-journal'
    assert citeproc['URL'] == 'https://www.ncbi.nlm.nih.gov/pubmed/21347133'
    assert citeproc['container-title'] == 'Summit on translational bioinformatics'
    assert citeproc['title'] == 'Secondary Use of EHR: Data Quality Issues and Informatics Opportunities.'
    assert citeproc['issued']['date-parts'] == [[2010, 3, 1]]
    authors = citeproc['author']
    assert authors[0]['given'] == 'Taxiarchis'
    assert authors[0]['family'] == 'Botsis'
    assert citeproc['PMID'] == '21347133'
    assert citeproc['PMCID'] == 'PMC3041534'


def test_citation_to_citeproc_pubmed_2():
    """
    Generated from XML returned by
    https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=pubmed&id=27094199&rettype=full
    """
    citation = 'pmid:27094199'
    citeproc = citation_to_citeproc(citation)
    print(citeproc)
    assert citeproc['id'] == 'alaFV9OY'
    assert citeproc['type'] == 'article-journal'
    assert citeproc['URL'] == 'https://www.ncbi.nlm.nih.gov/pubmed/27094199'
    assert citeproc['container-title'] == 'Circulation. Cardiovascular genetics'
    assert citeproc['container-title-short'] == 'Circ Cardiovasc Genet'
    assert citeproc['page'] == '179-84'
    assert citeproc['title'] == 'Genetic Association-Guided Analysis of Gene Networks for the Study of Complex Traits.'
    assert citeproc['issued']['date-parts'] == [[2016, 4]]
    authors = citeproc['author']
    assert authors[0]['given'] == 'Casey S'
    assert authors[0]['family'] == 'Greene'
    assert citeproc['PMID'] == '27094199'
    assert citeproc['DOI'] == '10.1161/circgenetics.115.001181'


def test_citation_to_citeproc_pubmed_with_numeric_month():
    """
    Generated from XML returned by
    https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=pubmed&id=29028984&rettype=full

    See https://github.com/manubot/manubot/issues/69
    """
    citation = 'pmid:29028984'
    citeproc = citation_to_citeproc(citation)
    print(citeproc)
    assert citeproc['issued']['date-parts'] == [[2018, 3, 15]]


def test_citation_to_citeproc_pubmed_book():
    """
    Extracting CSL metadata from books in PubMed is not supported.
    Logic not implemented to parse XML returned by
    https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=pubmed&id=29227604&rettype=full
    """
    with pytest.raises(NotImplementedError):
        citation_to_citeproc('pmid:29227604')


def test_citation_to_citeproc_isbn():
    csl_item = citation_to_citeproc('isbn:9780387950693')
    assert csl_item['type'] == 'book'
    assert csl_item['title'] == 'Complex analysis'




import copy
import functools
import logging

citeproc_type_fixer = {
    'journal-article': 'article-journal',
    'book-chapter': 'chapter',
    'posted-content': 'manuscript',
    'proceedings-article': 'paper-conference',
    'standard': 'entry',
    'reference-entry': 'entry',
}


def citeproc_passthrough(csl_item, set_id=None, prune=True):
    """
    Fix errors in a CSL item, according to the CSL JSON schema, and optionally
    change its id.

    http://docs.citationstyles.org/en/1.0.1/specification.html
    http://citeproc-js.readthedocs.io/en/latest/csl-json/markup.html
    https://github.com/citation-style-language/schema/blob/master/csl-data.json
    """
    if set_id is not None:
        csl_item['id'] = set_id
    logging.debug(f"Starting citeproc_passthrough with{'' if prune else 'out'} CSL pruning for id: {csl_item.get('id', 'id not specified')}")

    # Correct invalid CSL item types
    # See https://github.com/CrossRef/rest-api-doc/issues/187
    if 'type' in csl_item:
        csl_item['type'] = citeproc_type_fixer.get(csl_item['type'], csl_item['type'])

    if prune:
        # Remove fields that violate the CSL Item JSON Schema
        csl_item, = remove_jsonschema_errors([csl_item])

    # Default CSL type to entry
    csl_item['type'] = csl_item.get('type', 'entry')

    if prune:
        # Confirm that corrected CSL validates
        validator = get_jsonschema_csl_validator()
        validator.validate([csl_item])
    return csl_item


@functools.lru_cache()
def get_jsonschema_csl_validator():
    """
    Return a jsonschema validator for the CSL Item JSON Schema
    """
    import jsonref
    import jsonschema

    url = 'https://github.com/dhimmel/schema/raw/manubot/csl-data.json'
    # Use jsonref to workaround https://github.com/Julian/jsonschema/issues/447
    schema = jsonref.load_uri(url, jsonschema=True)
    Validator = jsonschema.validators.validator_for(schema)
    Validator.check_schema(schema)
    return Validator(schema)


def remove_jsonschema_errors(instance, recurse_depth=5):
    """
    Remove fields in CSL Items that produce JSON Schema errors. Should errors
    be removed, but the JSON instance still fails to validate, recursively call
    remove_jsonschema_errors until the instance validates or the recursion
    depth limit is reached.

    Note that this method may not be work for all types of JSON Schema errors
    and users looking to adapt it for other applications should write
    task-specific tests to provide empirical evaluate that it works as
    intended.

    See also:
    https://github.com/Julian/jsonschema/issues/448
    https://stackoverflow.com/questions/44694835
    """
    validator = get_jsonschema_csl_validator()
    errors = list(validator.iter_errors(instance))
    instance = copy.deepcopy(instance)
    errors = sorted(errors, key=lambda e: e.path, reverse=True)
    for error in errors:
        _remove_error(instance, error)
    if validator.is_valid(instance) or recurse_depth < 1:
        return instance
    return remove_jsonschema_errors(instance, recurse_depth - 1)


def _delete_elem(instance, path, absolute_path=None, message=''):
    """
    Helper function for remove_jsonschema_errors that deletes an element in the
    JSON-like input instance at the specified path. absolute_path is relative
    to the original validated instance for logging purposes. Defaults to path,
    if not specified. message is an optional string with additional error
    information to log.
    """
    if absolute_path is None:
        absolute_path = path
    logging.debug(
        (f'{message}\n' if message else message) +
        '_delete_elem deleting CSL element at: ' +
        '/'.join(map(str, absolute_path))
    )
    *head, tail = path
    try:
        del _deep_get(instance, head)[tail]
    except KeyError:
        pass


def _deep_get(instance, path):
    """
    Descend path to return a deep element in the JSON object instance.
    """
    for key in path:
        instance = instance[key]
    return instance


def _remove_error(instance, error):
    """
    Remove a jsonschema ValidationError from the JSON-like instance.

    See ValidationError documentation at
    http://python-jsonschema.readthedocs.io/en/latest/errors/#jsonschema.exceptions.ValidationError
    """
    sub_errors = error.context
    if sub_errors:
        # already_removed_additional was neccessary to workaround https://github.com/citation-style-language/schema/issues/154
        already_removed_additional = False
        for sub_error in sub_errors:
            if sub_error.validator == 'additionalProperties':
                if already_removed_additional:
                    continue
                already_removed_additional = True
            sub_instance = _deep_get(instance, error.path)
            _remove_error(sub_instance, sub_error)
    elif error.validator == 'additionalProperties':
        extras = set(error.instance) - set(error.schema['properties'])
        logging.debug(
            error.message +
            f'\nWill now remove these {len(extras)} additional properties.'
        )
        for key in extras:
            _delete_elem(
                instance=instance,
                path=list(error.path) + [key],
                absolute_path=list(error.absolute_path) + [key]
            )
    elif error.validator in {'enum', 'type', 'minItems', 'maxItems'}:
        _delete_elem(instance, error.path, error.absolute_path, error.message)
    elif error.validator == 'required':
        logging.warning(
            (f'{error.message}\n' if error.message else error.message) +
            'requried element missing at: ' +
            '/'.join(map(str, error.absolute_path))
        )
    else:
        raise NotImplementedError(f'{error.validator} is not yet supported')



import pytest

from manubot.cite.pubmed import (
    get_pmcid_and_pmid_for_doi,
    get_pmid_for_doi,
    get_pubmed_ids_for_doi,
)


@pytest.mark.parametrize(('doi', 'pmid'), [
    ('10.1098/rsif.2017.0387', '29618526'),  # in PubMed and PMC
    ('10.1161/CIRCGENETICS.115.001181', '27094199'),  # in PubMed but not PMC
    ('10.7717/peerj-cs.134', None),  # DOI in journal not indexed by PubMed
    ('10.1161/CIRC', None),  # invalid DOI
])
def test_get_pmid_for_doi(doi, pmid):
    output = get_pmid_for_doi(doi)
    assert pmid == output


@pytest.mark.parametrize(('doi', 'id_dict'), [
    ('10.1098/rsif.2017.0387', {'PMCID': 'PMC5938574', 'PMID': '29618526'}),
    ('10.7554/ELIFE.32822', {'PMCID': 'PMC5832410', 'PMID': '29424689'}),
    ('10.1161/CIRCGENETICS.115.001181', {}),  # only in PubMed, not in PMC
    ('10.7717/peerj.000', {}),  # Non-existent DOI
    ('10.peerj.000', {}),  # malformed DOI
])
def test_get_pmcid_and_pmid_for_doi(doi, id_dict):
    output = get_pmcid_and_pmid_for_doi(doi)
    assert id_dict == output


@pytest.mark.parametrize(('doi', 'id_dict'), [
    ('10.1098/rsif.2017.0387', {'PMCID': 'PMC5938574', 'PMID': '29618526'}),
    ('10.7554/ELIFE.32822', {'PMCID': 'PMC5832410', 'PMID': '29424689'}),
    ('10.1161/CIRCGENETICS.115.001181', {'PMID': '27094199'}),  # only in PubMed, not in PMC
    ('10.7717/peerj.000', {}),  # Non-existent DOI
])
def test_get_pubmed_ids_for_doi(doi, id_dict):
    output = get_pubmed_ids_for_doi(doi)
    assert id_dict == output



import collections
import functools
import json
import logging
import xml.etree.ElementTree

import requests

from manubot.util import get_manubot_user_agent


def get_pmc_citeproc(pmcid):
    """
    Get the citeproc JSON for a PubMed Central record by its PMID, PMCID, or
    DOI, using the NCBI Citation Exporter API.

    https://api.ncbi.nlm.nih.gov/lit/ctxp
    https://github.com/manubot/manubot/issues/21
    https://twitter.com/dhimmel/status/1061787168820092929
    """
    assert pmcid.startswith('PMC')
    csl_item = _get_literature_citation_exporter_csl_item('pmc', pmcid[3:])
    if 'URL' not in csl_item:
        csl_item['URL'] = f"https://www.ncbi.nlm.nih.gov/pmc/articles/{csl_item.get('PMCID', pmcid)}/"
    return csl_item


def _get_literature_citation_exporter_csl_item(database, identifier):
    """
    https://api.ncbi.nlm.nih.gov/lit/ctxp
    """
    if database not in {'pubmed', 'pmc'}:
        logging.error(
            f'Error calling _get_literature_citation_exporter_csl_item.\n'
            f'database must be either "pubmed" or "pmc", not {database}'
        )
        assert False
    if not identifier:
        logging.error(
            f'Error calling _get_literature_citation_exporter_csl_item.\n'
            f'identifier cannot be blank'
        )
        assert False
    params = {
        'format': 'csl',
        'id': identifier,
    }
    headers = {
        'User-Agent': get_manubot_user_agent(),
    }
    url = f'https://api.ncbi.nlm.nih.gov/lit/ctxp/v1/{database}/'
    response = requests.get(url, params, headers=headers)
    try:
        csl_item = response.json()
    except Exception as error:
        logging.error(
            f'Error fetching {database} metadata for {identifier}.\n'
            f'Invalid JSON response from {response.url}:\n{response.text}'
        )
        raise error
    assert isinstance(csl_item, dict)
    if csl_item.get('status', 'okay') == 'error':
        logging.error(
            f'Error fetching {database} metadata for {identifier}.\n'
            f'Literature Citation Exporter returned JSON indicating an error for {response.url}\n'
            f'{json.dumps(csl_item, indent=2)}'
        )
        assert False
    return csl_item


def get_pubmed_citeproc(pmid):
    """
    Query NCBI E-Utilities to create CSL Items for PubMed IDs.

    https://github.com/manubot/manubot/issues/21
    https://github.com/ncbi/citation-exporter/issues/3#issuecomment-355313143
    """
    pmid = str(pmid)
    params = {
        'db': 'pubmed',
        'id': pmid,
        'rettype': 'full',
    }
    headers = {
        'User-Agent': get_manubot_user_agent(),
    }
    url = 'https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi'
    with _get_eutils_rate_limiter():
        response = requests.get(url, params, headers=headers)
    try:
        element_tree = xml.etree.ElementTree.fromstring(response.text)
        element_tree, = list(element_tree)
    except Exception as error:
        logging.error(f'Error fetching PubMed metadata for {pmid}.\n'
                      f'Invalid XML response from {response.url}:\n{response.text}')
        raise error
    try:
        citeproc = citeproc_from_pubmed_article(element_tree)
    except Exception as error:
        msg = f'Error parsing the following PubMed metadata for PMID {pmid}:\n{response.text}'
        logging.error(msg)
        raise error
    return citeproc


def citeproc_from_pubmed_article(article):
    """
    article is a PubmedArticle xml element tree

    https://github.com/citation-style-language/schema/blob/master/csl-data.json
    """
    citeproc = collections.OrderedDict()

    if not article.find("MedlineCitation/Article"):
        raise NotImplementedError('Unsupported PubMed record: no <Article> element')

    title = article.findtext("MedlineCitation/Article/ArticleTitle")
    if title:
        citeproc['title'] = title

    volume = article.findtext("MedlineCitation/Article/Journal/JournalIssue/Volume")
    if volume:
        citeproc['volume'] = volume

    issue = article.findtext("MedlineCitation/Article/Journal/JournalIssue/Issue")
    if issue:
        citeproc['issue'] = issue

    page = article.findtext("MedlineCitation/Article/Pagination/MedlinePgn")
    if page:
        citeproc['page'] = page

    journal = article.findtext("MedlineCitation/Article/Journal/Title")
    if journal:
        citeproc['container-title'] = journal

    journal_short = article.findtext("MedlineCitation/Article/Journal/ISOAbbreviation")
    if journal_short:
        citeproc['container-title-short'] = journal_short

    issn = article.findtext("MedlineCitation/Article/Journal/ISSN")
    if issn:
        citeproc['ISSN'] = issn

    date_parts = extract_publication_date_parts(article)
    if date_parts:
        citeproc['issued'] = {'date-parts': [date_parts]}

    authors_csl = list()
    authors = article.findall("MedlineCitation/Article/AuthorList/Author")
    for author in authors:
        author_csl = collections.OrderedDict()
        given = author.findtext('ForeName')
        if given:
            author_csl['given'] = given
        family = author.findtext('LastName')
        if family:
            author_csl['family'] = family
        authors_csl.append(author_csl)
    if authors_csl:
        citeproc['author'] = authors_csl

    for id_type, key in ('pubmed', 'PMID'), ('pmc', 'PMCID'), ('doi', 'DOI'):
        xpath = f"PubmedData/ArticleIdList/ArticleId[@IdType='{id_type}']"
        value = article.findtext(xpath)
        if value:
            citeproc[key] = value.lower() if key == 'DOI' else value

    abstract = article.findtext("MedlineCitation/Article/Abstract/AbstractText")
    if abstract:
        citeproc['abstract'] = abstract

    citeproc['URL'] = f"https://www.ncbi.nlm.nih.gov/pubmed/{citeproc['PMID']}"
    citeproc['type'] = 'article-journal'

    return citeproc


month_abbrev_to_int = {
    'Jan': 1,
    'Feb': 2,
    'Mar': 3,
    'Apr': 4,
    'May': 5,
    'Jun': 6,
    'Jul': 7,
    'Aug': 8,
    'Sep': 9,
    'Oct': 10,
    'Nov': 11,
    'Dec': 12,
}


def extract_publication_date_parts(article):
    """
    Extract date published from a PubmedArticle xml element tree.
    """
    date_parts = []

    # Electronic articles
    date = article.find("MedlineCitation/Article/ArticleDate")
    if date:
        for part in 'Year', 'Month', 'Day':
            part = date.findtext(part)
            if not part:
                break
            date_parts.append(int(part))
        return date_parts

    # Print articles
    date = article.find("MedlineCitation/Article/Journal/JournalIssue/PubDate")
    year = date.findtext('Year')
    if year:
        date_parts.append(int(year))
    month = date.findtext('Month')
    if month:
        try:
            date_parts.append(month_abbrev_to_int[month])
        except KeyError:
            date_parts.append(int(month))
    day = date.findtext('Day')
    if day:
        date_parts.append(int(day))
    return date_parts


def get_pmcid_and_pmid_for_doi(doi):
    """
    Query PMC's ID Converter API to retrieve the PMCID and PMID for a DOI.
    Does not work for DOIs that are in Pubmed but not PubMed Central.
    https://www.ncbi.nlm.nih.gov/pmc/tools/id-converter-api/
    """
    assert isinstance(doi, str)
    assert doi.startswith('10.')
    params = {
        'ids': doi,
        'tool': 'manubot',
    }
    url = 'https://www.ncbi.nlm.nih.gov/pmc/utils/idconv/v1.0/'
    response = requests.get(url, params)
    if not response.ok:
        logging.warning(f'Status code {response.status_code} querying {response.url}\n')
        return {}
    try:
        element_tree = xml.etree.ElementTree.fromstring(response.text)
    except Exception:
        logging.warning(f'Error fetching PMC ID conversion for {doi}.\n'
                        f'Response from {response.url}:\n{response.text}')
        return {}
    records = element_tree.findall('record')
    if len(records) != 1:
        logging.warning(f'Expected PubMed Central ID converter to return a single XML record for {doi}.\n'
                        f'Response from {response.url}:\n{response.text}')
        return {}
    record, = records
    if record.findtext('status', default='okay') == 'error':
        return {}
    id_dict = {}
    for id_type in 'pmcid', 'pmid':
        id_ = record.get(id_type)
        if id_:
            id_dict[id_type.upper()] = id_
    return id_dict


def get_pmid_for_doi(doi):
    """
    Query NCBI's E-utilities to retrieve the PMID for a DOI.
    """
    assert isinstance(doi, str)
    assert doi.startswith('10.')
    params = {
        'db': 'pubmed',
        'term': f'{doi}[DOI]',
    }
    headers = {
        'User-Agent': get_manubot_user_agent(),
    }
    url = 'https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi'
    with _get_eutils_rate_limiter():
        response = requests.get(url, params, headers=headers)
    if not response.ok:
        logging.warning(f'Status code {response.status_code} querying {response.url}\n')
        return None
    try:
        element_tree = xml.etree.ElementTree.fromstring(response.text)
    except Exception as error:
        logging.warning(f'Error in ESearch XML for DOI: {doi}.\n'
                        f'Response from {response.url}:\n{response.text}')
        return None
    id_elems = element_tree.findall('IdList/Id')
    if len(id_elems) != 1:
        logging.debug(f'No PMIDs found for {doi}.\n'
                      f'Response from {response.url}:\n{response.text}')
        return None
    id_elem, = id_elems
    return id_elem.text


def get_pubmed_ids_for_doi(doi):
    """
    Return a dictionary with PMCID and PMID, if they exist, for the specified
    DOI. See https://github.com/manubot/manubot/issues/45.
    """
    pubmed_ids = get_pmcid_and_pmid_for_doi(doi)
    if not pubmed_ids:
        pmid = get_pmid_for_doi(doi)
        if pmid:
            pubmed_ids['PMID'] = pmid
    return pubmed_ids

@functools.lru_cache()
def _get_eutils_rate_limiter():
    """
    Rate limiter to cap NCBI E-utilities queries to 3 per second as per
    https://ncbiinsights.ncbi.nlm.nih.gov/2017/11/02/new-api-keys-for-the-e-utilities/
    """
    from ratelimiter import RateLimiter
    return RateLimiter(max_calls=2, period=1)



"""
Manubot's command line interface
"""
import argparse
import logging
import pathlib
import sys
import warnings

import manubot
from manubot.util import import_function


def parse_arguments():
    """
    Read and process command line arguments.
    """
    parser = argparse.ArgumentParser(description='Manubot: the manuscript bot for scholarly writing')
    parser.add_argument('--version', action='version', version=f'v{manubot.__version__}')
    subparsers = parser.add_subparsers(
        title='subcommands',
        description='All operations are done through subcommands:',
    )
    # Require specifying a sub-command
    subparsers.required = True  # https://bugs.python.org/issue26510
    subparsers.dest = 'subcommand'  # https://bugs.python.org/msg186387
    add_subparser_process(subparsers)
    add_subparser_cite(subparsers)
    for subparser in subparsers.choices.values():
        subparser.add_argument(
            '--log-level',
            default='WARNING',
            choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
            help='Set the logging level for stderr logging',
        )
    args = parser.parse_args()
    return args


def add_subparser_process(subparsers):
    parser = subparsers.add_parser(
        name='process', help='process manuscript content',
        description='Process manuscript content to create outputs for Pandoc consumption. '
                    'Performs bibliographic processing and templating.',
    )
    parser.add_argument(
        '--content-directory', type=pathlib.Path, required=True,
        help='Directory where manuscript content files are located.',
    )
    parser.add_argument(
        '--output-directory', type=pathlib.Path, required=True,
        help='Directory to output files generated by this script.',
    )
    parser.add_argument(
        '--template-variables-path', action='append', default=[],
        help='Path or URL of a JSON file containing template variables for jinja2. '
             'Specify this argument multiple times to read multiple files. '
             'Variables can be applied to a namespace (i.e. stored under a dictionary key) '
             'like `--template-variables-path=namespace=path_or_url`. '
             'Namespaces must match the regex `[a-zA-Z_][a-zA-Z0-9_]*`.',
    )
    parser.add_argument(
        '--cache-directory', type=pathlib.Path,
        help='Custom cache directory. '
             'If not specified, caches to output-directory.',
    )
    parser.add_argument(
        '--clear-requests-cache', action='store_true',
    )
    parser.set_defaults(function='manubot.process.process_command.cli_process')


def add_subparser_cite(subparsers):
    parser = subparsers.add_parser(
        name='cite',
        help='citation to CSL command line utility',
        description='Retrieve bibliographic metadata for one or more citation identifiers.',
    )
    parser.add_argument(
        '--render',
        action='store_true',
        help='Whether to render CSL Data into a formatted reference list using Pandoc. '
             'Pandoc version 2.0 or higher is required for complete support of available output formats.',
    )
    parser.add_argument(
        '--csl',
        default='https://github.com/greenelab/manubot-rootstock/raw/master/build/assets/style.csl',
        help="When --render, specify an XML CSL definition to style references (i.e. Pandoc's --csl option). "
             "Defaults to Manubot's style.",
    )
    parser.add_argument(
        '--format',
        choices=['plain', 'markdown', 'docx', 'html', 'jats'],
        help="When --render, format to use for output file. "
             "If not specified, attempt to infer this from filename extension. "
             "Otherwise, default to plain.",
    )
    parser.add_argument(
        '--output',
        type=pathlib.Path,
        help='Specify a file to write output, otherwise default to stdout.',
    )
    parser.add_argument(
        '--allow-invalid-csl-data',
        dest='prune_csl',
        action='store_false',
        help='Allow CSL Items that do not conform to the JSON Schema. Skips CSL pruning.',
    )
    parser.add_argument(
        'citations',
        nargs='+',
        help='One or more (space separated) citations to produce CSL for.',
    )
    parser.set_defaults(function='manubot.cite.cite_command.cli_cite')


def main():
    """
    Called as a console_scripts entry point in setup.py. This function defines
    the manubot command line script.
    """
    # Track if message gets logged with severity of error or greater
    # See https://stackoverflow.com/a/45446664/4651668
    import errorhandler
    error_handler = errorhandler.ErrorHandler()

    # Log DeprecationWarnings
    warnings.simplefilter('always', DeprecationWarning)
    logging.captureWarnings(True)

    # Log to stderr
    logger = logging.getLogger()
    stream_handler = logging.StreamHandler(stream=sys.stderr)
    stream_handler.setFormatter(logging.Formatter('## {levelname}\n{message}', style='{'))
    logger.addHandler(stream_handler)

    args = parse_arguments()
    logger.setLevel(getattr(logging, args.log_level))

    function = import_function(args.function)
    function(args)

    if error_handler.fired:
        logging.critical('Failure: exiting with code 1 due to logged errors')
        raise SystemExit(1)



import json
import logging
import re


def get_url_citeproc(url):
    """
    Get citeproc for a URL trying a sequence of strategies.

    This function uses a list of CSL JSON Item metadata retrievers, specified
    by the module-level variable `url_retrievers`. The methods are attempted
    in order, with this function returning the metadata from the first
    non-failing method.
    """
    for retriever in url_retrievers:
        try:
            return retriever(url)
        except Exception as error:
            logging.warning(
                f'Error in {retriever.__name__} for {url} '
                f'due to a {error.__class__.__name__}:\n{error}'
            )
            logging.info(error, exc_info=True)
    raise Exception(f'all get_url_citeproc methods failed for {url}')


def get_url_citeproc_zotero(url):
    """
    Use Zotero's translation-server to generate a CSL Item for the specified URL.
    """
    from manubot.cite.zotero import (
        export_as_csl,
        web_query,
    )
    zotero_data = web_query(url)
    csl_data = export_as_csl(zotero_data)
    csl_item, = csl_data
    return csl_item


def get_url_citeproc_greycite(url):
    """
    Uses Greycite which has experiened uptime problems in the past.
    API calls seem to take at least 15 seconds. Browser requests are much
    faster. Setting header did not have an effect. Consider mimicking browser
    using selenium.

    More information on Greycite at:
    http://greycite.knowledgeblog.org/
    http://knowledgeblog.org/greycite
    https://arxiv.org/abs/1304.7151
    https://git.io/v9N2C
    """
    import requests
    from manubot.util import get_manubot_user_agent
    headers = {
        'Connection': 'close',  # https://github.com/kennethreitz/requests/issues/4023
        'User-Agent': get_manubot_user_agent(),
    }
    response = requests.get(
        'http://greycite.knowledgeblog.org/json',
        params={'uri': url},
        headers=headers,
    )
    # Some Greycite responses were valid JSON besides for an error appended
    # like "<p>*** Date set from uri<p>" or "<p>*** fetch error : 404<p>".
    pattern = re.compile(r"<p>\*\*\*.*<p>")
    text = pattern.sub('', response.text)
    csl_item = json.loads(text)
    csl_item['type'] = 'webpage'
    return csl_item


def get_url_citeproc_manual(url):
    """
    Manually create citeproc for a URL.
    """
    return {
        'URL': url,
        'type': 'webpage',
    }


url_retrievers = [
    get_url_citeproc_zotero,
    get_url_citeproc_greycite,
    get_url_citeproc_manual,
]




def test_imports():
    import manubot.cite
    import manubot.cite.util
    import manubot.cite.arxiv
    import manubot.cite.doi
    import manubot.cite.pubmed
    import manubot.cite.url
    import manubot.process.manuscript
    assert isinstance(manubot.cite.util.citeproc_retrievers, dict)



from manubot.cite.url import get_url_citeproc_zotero


def test_get_url_citeproc_zotero_nyt():
    """
    This command creates two translation-server queries. The first query is
    equivalent to:
    ```
    curl --verbose \
      --header "Content-Type: text/plain" \
      --data 'https://nyti.ms/1NuB0WJ' \
      'https://translate.manubot.org/web'
    ```
    """
    url = 'https://nyti.ms/1NuB0WJ'
    csl_item = get_url_citeproc_zotero(url)
    assert csl_item['title'] == 'Unraveling the Ties of Altitude, Oxygen and Lung Cancer'
    assert csl_item['author'][0]['family'] == 'Johnson'


def test_get_url_citeproc_zotero_manubot():
    """
    This command creates two translation-server queries. The first query is
    equivalent to:
    ```
    curl --verbose \
      --header "Content-Type: text/plain" \
      --data 'https://greenelab.github.io/meta-review/v/0770300e1d5490a1ae8ff3a85ddca2cdc4ae0613/' \
      'https://translate.manubot.org/web'
    ```
    """
    url = 'https://greenelab.github.io/meta-review/v/0770300e1d5490a1ae8ff3a85ddca2cdc4ae0613/'
    csl_item = get_url_citeproc_zotero(url)
    assert csl_item['title'] == 'Open collaborative writing with Manubot'
    assert csl_item['author'][1]['family'] == 'Slochower'
    # Zotero CSL exporter returns mixed string/int date-parts
    # https://github.com/zotero/zotero/issues/1603
    assert [int(x) for x in csl_item['issued']['date-parts'][0]] == [2018, 12, 18]


def test_get_url_citeproc_zotero_github():
    """
    This command creates two translation-server queries. The first query is
    equivalent to:
    ```
    curl --verbose \
      --header "Content-Type: text/plain" \
      --data 'https://github.com/pandas-dev/pandas/tree/d5e5bf761092c59eeb9b8750f05f2bc29fb45927' \
      'https://translate.manubot.org/web'
    ```
    """
    url = 'https://github.com/pandas-dev/pandas/tree/d5e5bf761092c59eeb9b8750f05f2bc29fb45927'
    csl_item = get_url_citeproc_zotero(url)
    assert csl_item['title'].startswith('Flexible and powerful data analysis')
    assert csl_item['source'] == 'GitHub'



"""
Methods to interact with a Zotero translation-server.
https://github.com/zotero/translation-server

The Manubot team currently hosts a public translation server at
https://translate.manubot.org. More information on this instance at
https://github.com/manubot/manubot/issues/82.
"""

import json
import logging

import requests

from manubot.util import get_manubot_user_agent

base_url = 'https://translate.manubot.org'


def web_query(url):
    """
    Return Zotero citation metadata for a URL as a list containing a single element that
    is a dictionary with the URL's metadata.
    """
    headers = {
        'User-Agent': get_manubot_user_agent(),
        'Content-Type': 'text/plain',
    }
    params = {
        'single': 1,
    }
    api_url = f'{base_url}/web'
    response = requests.post(api_url, params=params, headers=headers, data=str(url))
    try:
        zotero_data = response.json()
    except Exception as error:
        logging.warning(f'Error parsing web_query output as JSON for {url}:\n{response.text}')
        raise error
    if response.status_code == 300:
        # When single=1 is specified, multiple results should never be returned
        logging.warning(
            f'web_query returned multiple results for {url}:\n' +
            json.dumps(zotero_data, indent=2)
        )
        raise ValueError(f'multiple results for {url}')
    zotero_data = _passthrough_zotero_data(zotero_data)
    return zotero_data


def search_query(identifier):
    """
    Supports DOI, ISBN, PMID, arXiv ID.
    curl -d 10.2307/4486062 -H 'Content-Type: text/plain' http://127.0.0.1:1969/search
    """
    api_url = f'{base_url}/search'
    headers = {
        'User-Agent': get_manubot_user_agent(),
        'Content-Type': 'text/plain',
    }
    response = requests.post(api_url, headers=headers, data=str(identifier))
    try:
        zotero_data = response.json()
    except Exception as error:
        logging.warning(f'Error parsing search_query output as JSON for {identifier}:\n{response.text}')
        raise error
    zotero_data = _passthrough_zotero_data(zotero_data)
    return zotero_data


def _passthrough_zotero_data(zotero_data):
    """
    Address known issues with Zotero metadata.
    Assumes zotero data should contain a single bibliographic record.
    """
    if not isinstance(zotero_data, list):
        raise ValueError('_passthrough_zotero_data: zotero_data should be a list')
    if len(zotero_data) > 1:
        # Sometimes translation-server creates multiple data items for a single record.
        # If so, keep only the parent item, and remove child items (such as notes).
        # https://github.com/zotero/translation-server/issues/67
        zotero_data = zotero_data[:1]
    return zotero_data


def export_as_csl(zotero_data):
    """
    Export Zotero JSON data to CSL JSON using a translation-server /export query.
    Performs a similar query to the following curl command:
    ```
    curl --verbose \
      --data @items.json \
      --header 'Content-Type: application/json' \
      'https://translate.manubot.org/export?format=csljson'
    ```
    """
    api_url = f'{base_url}/export'
    params = {
        'format': 'csljson',
    }
    headers = {
        'User-Agent': get_manubot_user_agent(),
    }
    response = requests.post(api_url, params=params, headers=headers, json=zotero_data)
    if not response.ok:
        message = f'export_as_csl: translation-server returned status code {response.status_code}'
        logging.warning(f'{message} with the following output:\n{response.text}')
        raise requests.HTTPError(message)
    try:
        csl_json = response.json()
    except Exception as error:
        logging.warning(f'Error parsing export_as_csl output as JSON:\n{response.text}')
        raise error
    return csl_json



import collections
import json
import logging
import os
import pathlib
import re
import textwrap
import warnings

import jinja2
import pandas
import requests
import requests_cache
import yaml

from manubot.process.manuscript import (
    datetime_now,
    get_citation_strings,
    get_manuscript_stats,
    get_text,
    replace_citations_strings_with_ids,
)
from manubot.cite.util import (
    citation_to_citeproc,
    get_citation_id,
    is_valid_citation_string,
    standardize_citation,
)

from manubot.cite.citeproc import citeproc_passthrough


def read_manual_references(path):
    """
    Read manual references (overrides) in JSON CSL specified by a pathlib.Path.
    Returns a standard_citation to citeproc dictionary.
    """
    if not path.is_file():
        return dict()
    with path.open() as read_file:
        csl_items = json.load(read_file)
    manual_refs = dict()
    for csl_item in csl_items:
        standard_citation = csl_item.pop('standard_citation')
        csl_item = citeproc_passthrough(csl_item, set_id=get_citation_id(standard_citation))
        manual_refs[standard_citation] = csl_item
    return manual_refs


def check_collisions(citation_df):
    """
    Check for citation_id hash collisions
    """
    collision_df = citation_df[['standard_citation', 'citation_id']].drop_duplicates()
    collision_df = collision_df[collision_df.citation_id.duplicated(keep=False)]
    if not collision_df.empty:
        logging.error(f'OMF! Hash collision. Congratulations.\n{collision_df}')
    return collision_df


def check_multiple_citation_strings(citation_df):
    """
    Identify different citation strings referring the the same reference.
    """
    message = textwrap.dedent(f'''\
    {len(citation_df)} unique citations strings extracted from text
    {citation_df.standard_citation.nunique()} unique standard citations\
    ''')
    logging.info(message)
    multi_df = citation_df[citation_df.standard_citation.duplicated(keep=False)]
    if not multi_df.empty:
        table = multi_df.to_string(index=False, columns=['standard_citation', 'string'])
        logging.warning(f'Multiple citation strings detected for the same reference:\n{table}')
    return multi_df


def read_json(path):
    """
    Read json from a path or URL.
    """
    if re.match('^(http|ftp)s?://', path):
        response = requests.get(path)
        obj = response.json(object_pairs_hook=collections.OrderedDict)
    else:
        path = pathlib.Path(path)
        with path.open() as read_file:
            obj = json.load(read_file, object_pairs_hook=collections.OrderedDict)
    return obj


def read_jsons(paths):
    """
    Read multiple JSON files into a user_variables dictionary. Provide a list
    of paths (URLs or filepaths). Paths can optionally have a namespace
    prepended. For example:

    ```
    paths = [
        'https://git.io/vbkqm',  # update the dictionary's top-level
        'namespace_1=https://git.io/vbkqm',  # store under 'namespace_1' key
        'namespace_2=some_local_path.json',  # store under 'namespace_2' key
    ]
    ```

    If a namespace is not provided, the JSON must contain a dictionary as its
    top level. Namespaces should consist only of ASCII alphanumeric characters
    (includes underscores, first character cannot be numeric).
    """
    user_variables = collections.OrderedDict()
    for path in paths:
        logging.info(f'Read the following user-provided templating variables for {path}')
        # Match only namespaces that are valid jinja2 variable names
        # http://jinja.pocoo.org/docs/2.10/api/#identifier-naming
        match = re.match(r'([a-zA-Z_][a-zA-Z0-9_]*)=(.+)', path)
        if match:
            namespace, path = match.groups()
            logging.info(f'Using the "{namespace}" namespace for template variables from {path}')
        try:
            obj = read_json(path)
        except Exception:
            logging.exception(f'Error reading template variables from {path}')
            continue
        if match:
            obj = {namespace: obj}
        assert isinstance(obj, dict)
        conflicts = user_variables.keys() & obj.keys()
        if conflicts:
            logging.warning(f'Template variables in {path} overwrite existing '
                            'values for the following keys:\n' +
                            '\n'.join(conflicts))
        user_variables.update(obj)
    logging.info(f'Reading user-provided templating variables complete:\n'
                 f'{json.dumps(user_variables, indent=2, ensure_ascii=False)}')
    return user_variables


def add_author_affiliations(variables):
    """
    Edit variables to contain numbered author affiliations. Specifically,
    add a list of affiliation_numbers for each author and add a list of
    affiliations to the top-level of variables. If no authors have any
    affiliations, variables is left unmodified.
    """
    rows = list()
    for author in variables['authors']:
        if 'affiliations' not in author:
            continue
        if not isinstance(author['affiliations'], list):
            warnings.warn(
                f"Expected list for {author['name']}'s affiliations. "
                f"Assuming multiple affiliations are `; ` separated. "
                f"Please switch affiliations to a list.",
                category=DeprecationWarning
            )
            author['affiliations'] = author['affiliations'].split('; ')
        for affiliation in author['affiliations']:
            rows.append((author['name'], affiliation))
    if not rows:
        return variables
    affil_map_df = pandas.DataFrame(rows, columns=['name', 'affiliation'])
    affiliation_df = affil_map_df[['affiliation']].drop_duplicates()
    affiliation_df['affiliation_number'] = range(1, 1 + len(affiliation_df))
    affil_map_df = affil_map_df.merge(affiliation_df)
    name_to_numbers = {name: sorted(df.affiliation_number) for name, df in
                       affil_map_df.groupby('name')}
    for author in variables['authors']:
        author['affiliation_numbers'] = name_to_numbers.get(author['name'], [])
    variables['affiliations'] = affiliation_df.to_dict(orient='records')
    return variables


def get_metadata_and_variables(args):
    """
    Process metadata.yaml and create variables available for jinja2 templating.
    """
    # Generated manuscript variables
    variables = collections.OrderedDict()

    # Read metadata which contains pandoc_yaml_metadata
    # as well as author_info.
    if args.meta_yaml_path.is_file():
        with args.meta_yaml_path.open() as read_file:
            metadata = yaml.load(read_file)
            assert isinstance(metadata, dict)
    else:
        metadata = {}
        logging.warning(f'missing {args.meta_yaml_path} file with yaml_metadata_block for pandoc')

    # Add date to metadata
    now = datetime_now()
    logging.info(
        f'Using {now:%Z} timezone.\n'
        f'Dating manuscript with the current datetime: {now.isoformat()}')
    metadata['date-meta'] = now.date().isoformat()
    variables['date'] = f'{now:%B} {now.day}, {now.year}'

    # Process authors metadata
    authors = metadata.pop('author_info', [])
    if authors is None:
        authors = []
    metadata['author-meta'] = [author['name'] for author in authors]
    variables['authors'] = authors
    variables = add_author_affiliations(variables)

    # Set repository version metadata for Travis CI builds only
    # https://docs.travis-ci.com/user/environment-variables/
    if os.getenv('TRAVIS', 'false') == 'true':
        repo_slug = os.environ['TRAVIS_REPO_SLUG']
        repo_owner, repo_name = repo_slug.split('/')
        variables['ci_source'] = {
            'repo_slug': repo_slug,
            'repo_owner': repo_owner,
            'repo_name': repo_name,
            'commit': os.environ['TRAVIS_COMMIT'],
        }

    # Update variables with user-provided variables here
    user_variables = read_jsons(args.template_variables_path)
    variables.update(user_variables)

    return metadata, variables


def get_citation_df(args, text):
    """
    Generate citation_df and save it to 'citations.tsv'.
    """
    citation_df = pandas.DataFrame(
        {'string': get_citation_strings(text)}
    )
    if args.citation_tags_path.is_file():
        tag_df = pandas.read_csv(args.citation_tags_path, sep='\t')
        na_rows_df = tag_df[tag_df.isnull().any(axis='columns')]
        if not na_rows_df.empty:
            logging.error(
                f'{args.citation_tags_path} contains rows with missing values:\n'
                f'{na_rows_df}\n'
                'This error can be caused by using spaces rather than tabs to delimit fields.\n'
                'Proceeding to reread TSV with delim_whitespace=True.'
            )
            tag_df = pandas.read_csv(args.citation_tags_path, delim_whitespace=True)
        tag_df['string'] = '@tag:' + tag_df.tag
        for citation in tag_df.citation:
            is_valid_citation_string('@' + citation, allow_raw=True)
        citation_df = citation_df.merge(tag_df[['string', 'citation']], how='left')
    else:
        citation_df['citation'] = None
        logging.info(f'missing {args.citation_tags_path} file: no citation tags set')
    citation_df.citation.fillna(citation_df.string.astype(str).str.lstrip('@'), inplace=True)
    citation_df['standard_citation'] = citation_df.citation.map(standardize_citation)
    citation_df['citation_id'] = citation_df.standard_citation.map(get_citation_id)
    citation_df = citation_df.sort_values(['standard_citation', 'citation'])
    citation_df.to_csv(args.citations_path, sep='\t', index=False)
    check_collisions(citation_df)
    check_multiple_citation_strings(citation_df)
    return citation_df


def generate_csl_items(args, citation_df):
    """
    General CSL (citeproc) items for standard_citations in citation_df.
    Writes references.json to disk and logs warnings for potential problems.
    """
    # Read manual references (overrides) in JSON CSL
    manual_refs = read_manual_references(args.manual_references_path)

    requests_cache.install_cache(args.requests_cache_path, include_get_headers=True)
    cache = requests_cache.get_cache()
    if args.clear_requests_cache:
        logging.info('Clearing requests-cache')
        requests_cache.clear()
    logging.info(f'requests-cache starting with {len(cache.responses)} cached responses')

    csl_items = list()
    failures = list()
    for citation in citation_df.standard_citation.unique():
        if citation in manual_refs:
            csl_items.append(manual_refs[citation])
            continue
        elif citation.startswith('raw:'):
            logging.error(
                f'CSL JSON Data with a standard_citation of {citation} not found in manual-references.json. '
                'Metadata must be provided for raw citations.'
            )
            failures.append(citation)
        try:
            citeproc = citation_to_citeproc(citation)
            csl_items.append(citeproc)
        except Exception as error:
            logging.exception(f'Citeproc retrieval failure for {citation}')
            failures.append(citation)

    logging.info(f'requests-cache finished with {len(cache.responses)} cached responses')
    requests_cache.uninstall_cache()

    if failures:
        message = 'CSL JSON Data retrieval failed for:\n{}'.format(
            '\n'.join(failures))
        logging.error(message)

    # Write JSON CSL bibliography for Pandoc.
    with args.references_path.open('w') as write_file:
        json.dump(csl_items, write_file, indent=2, ensure_ascii=False)
        write_file.write('\n')
    return csl_items


def template_with_jinja2(text, variables):
    """
    Template using jinja2 with the variables dictionary unpacked as keyword
    arguments.
    """
    jinja_environment = jinja2.Environment(
        loader=jinja2.BaseLoader(),
        undefined=jinja2.make_logging_undefined(logging.getLogger()),
        comment_start_string='{##',
        comment_end_string='##}',
    )
    template = jinja_environment.from_string(text)
    return template.render(**variables)


def prepare_manuscript(args):
    """
    Compile manuscript, creating manuscript.md and references.json as inputs
    for pandoc.
    """
    text = get_text(args.content_directory)
    citation_df = get_citation_df(args, text)

    generate_csl_items(args, citation_df)

    citation_string_to_id = collections.OrderedDict(
        zip(citation_df.string, citation_df.citation_id))
    text = replace_citations_strings_with_ids(text, citation_string_to_id)

    metadata, variables = get_metadata_and_variables(args)
    variables['manuscript_stats'] = get_manuscript_stats(text, citation_df)
    with args.variables_path.open('w') as write_file:
        json.dump(variables, write_file, indent=2)
        write_file.write('\n')

    text = template_with_jinja2(text, variables)

    # Write manuscript for pandoc
    with args.manuscript_path.open('w') as write_file:
        yaml.dump(metadata, write_file, default_flow_style=False,
                  explicit_start=True, explicit_end=True)
        write_file.write('\n')
        write_file.write(text)



import collections
import json
import logging
import re


def get_isbn_citeproc(isbn):
    """
    Generate CSL JSON Data for an ISBN. Converts all ISBNs to 13-digit format.

    This function uses a list of CSL JSON Item metadata retrievers, specified
    by the module-level variable `isbn_retrievers`. The methods are attempted
    in order, with this function returning the metadata from the first
    non-failing method.
    """
    import isbnlib
    isbn = isbnlib.to_isbn13(isbn)
    for retriever in isbn_retrievers:
        try:
            return retriever(isbn)
        except Exception as error:
            logging.warning(
                f'Error in {retriever.__name__} for {isbn} '
                f'due to a {error.__class__.__name__}:\n{error}'
            )
            logging.info(error, exc_info=True)
    raise Exception(f'all get_isbn_citeproc methods failed for {isbn}')


def get_isbn_citeproc_zotero(isbn):
    """
    Generate CSL JSON Data for an ISBN using Zotero's translation-server.
    """
    from manubot.cite.zotero import export_as_csl, search_query
    zotero_data = search_query(f'isbn:{isbn}')
    csl_data = export_as_csl(zotero_data)
    csl_item, = csl_data
    return csl_item


def get_isbn_citeproc_citoid(isbn):
    """
    Return CSL JSON Data for an ISBN using the Wikipedia Citoid API.
    https://en.wikipedia.org/api/rest_v1/#!/Citation/getCitation
    """
    import requests
    from manubot.util import get_manubot_user_agent
    headers = {
        'User-Agent': get_manubot_user_agent(),
    }
    url = f'https://en.wikipedia.org/api/rest_v1/data/citation/mediawiki/{isbn}'
    response = requests.get(url, headers=headers)
    result = response.json()
    if isinstance(result, dict):
        if result['title'] == 'Not found.':
            raise KeyError(f'Metadata for ISBN {isbn} not found at {url}')
        else:
            raise Exception(
                f'Unable to extract CSL from JSON metadata for ISBN {isbn}:\n'
                f'{json.dumps(result.text)}'
            )
    mediawiki, = result
    csl_data = collections.OrderedDict()
    csl_data['type'] = mediawiki.get('itemType', 'book')
    if 'title' in mediawiki:
        csl_data['title'] = mediawiki['title']
    if 'author' in mediawiki:
        csl_author = list()
        for last, first in mediawiki['author']:
            csl_author.append({
                'given': first,
                'family': last,
            })
        if csl_author:
            csl_data['author'] = csl_author
    if 'date' in mediawiki:
        year_pattern = re.compile(r'[0-9]{4}')
        match = year_pattern.search(mediawiki['date'])
        if match:
            year = int(match.group())
            csl_data['issued'] = {'date-parts': [[year]]}
        else:
            logging.debug(
                f'get_isbn_citeproc_citoid: issue extracting date for ISBN {isbn}\n'
                f'metadata retrieved from {url}\n'
                f'unable to extract year from date field: {mediawiki["date"]}'
            )
    if 'publisher' in mediawiki:
        csl_data['publisher'] = mediawiki['publisher']
    if 'place' in mediawiki:
        csl_data['publisher-place'] = mediawiki['place']
    if 'volume' in mediawiki:
        csl_data['volume'] = mediawiki['volume']
    if 'edition' in mediawiki:
        csl_data['edition'] = mediawiki['edition']
    if 'abstractNote' in mediawiki:
        csl_data['abstract'] = mediawiki['abstractNote']
    csl_data['ISBN'] = isbn
    if 'source' in mediawiki:
        csl_data['source'] = mediawiki['source'][0]
    if 'url' in mediawiki:
        csl_data['URL'] = mediawiki['url']
    return csl_data


def get_isbn_citeproc_isbnlib(isbn):
    """
    Generate CSL JSON Data for an ISBN using isbnlib.
    """
    import isbnlib
    metadata = isbnlib.meta(isbn, cache=None)
    csl_json = isbnlib.registry.bibformatters['csl'](metadata)
    csl_data = json.loads(csl_json)
    return csl_data


isbn_retrievers = [
    get_isbn_citeproc_zotero,
    get_isbn_citeproc_citoid,
    get_isbn_citeproc_isbnlib,
]



import pathlib

import pytest

from manubot.process.util import (
    add_author_affiliations,
    read_jsons,
)

directory = pathlib.Path(__file__).parent.resolve()


def test_read_jsons_empty():
    paths = []
    user_variables = read_jsons(paths)
    assert isinstance(user_variables, dict) and not user_variables


def test_read_jsons():
    """
    Test reading multiple JSON files, from both local paths and URLs.
    """
    local_path = 'manuscripts/variables/content/template-variables.json'
    local_path = directory.joinpath(local_path)
    paths = [
        'https://git.io/vbkqm',
        'https://git.io/vbkqm',
        'namespace_1=https://git.io/vbkqm',
        'namespace_2=https://git.io/vbkqm',
        f'namespace_2={local_path}',
        f'namespace_3={local_path}',
    ]
    user_variables = read_jsons(paths)
    assert 'namespace_1' in user_variables
    assert 'namespace_2' in user_variables
    assert 'namespace_3' in user_variables
    assert user_variables['generated_by'] == 'Manubot'
    assert 'violet' in user_variables['namespace_1']['rainbow']
    assert 'yellow' in user_variables['namespace_2']['rainbow']
    assert 'orange' in user_variables['namespace_3']['rainbow']


def test_add_author_affiliations_empty():
    variables = {}
    variables['authors'] = [
        {'name': 'Jane Roe'},
        {'name': 'John Doe'},
    ]
    returned_variables = add_author_affiliations(variables)
    assert variables is returned_variables
    assert 'affiliations' not in variables
    for author in variables['authors']:
        assert 'affiliation_numbers' not in author


def test_add_author_affiliations():
    variables = {}
    variables['authors'] = [
        # Deprecated affiliations format (as a string that's `; ` separated)
        {'name': 'Jane Roe', 'affiliations': 'Department of Doe, University of Roe; Peppertea University'},
        # Prefered affiliations format as a list
        {'name': 'John Doe', 'affiliations': ['Unique University', 'Peppertea University']},
    ]
    with pytest.warns(DeprecationWarning):
        returned_variables = add_author_affiliations(variables)
    assert variables is returned_variables
    assert variables['affiliations'] == [
        {'affiliation': 'Department of Doe, University of Roe', 'affiliation_number': 1},
        {'affiliation': 'Peppertea University', 'affiliation_number': 2},
        {'affiliation': 'Unique University', 'affiliation_number': 3},
    ]
    authors = variables['authors']
    assert authors[0]['affiliations'] == [
        'Department of Doe, University of Roe',
        'Peppertea University',
    ]
    assert authors[0]['affiliation_numbers'] == [1, 2]
    assert authors[1]['affiliation_numbers'] == [2, 3]



import pathlib
import subprocess

import pytest

directory = pathlib.Path(__file__).parent.resolve()

# List of manuscripts for testing. All subdirectories of ./manuscripts
manuscripts = [path.name for path in
               directory.joinpath('manuscripts').iterdir() if path.is_dir()]


@pytest.mark.parametrize("manuscript", manuscripts)
def test_example_manuscript(manuscript):
    """
    Test command line execution of manubot to build an example manuscript.
    """
    manuscript_dir = directory.joinpath('manuscripts', manuscript)
    args = [
        'manubot',
        'process',
        '--log-level', 'INFO',
        '--content-directory', str(manuscript_dir.joinpath('content')),
        '--output-directory', str(manuscript_dir.joinpath('output')),
    ]
    if manuscript == 'variables':
        args.extend([
            '--template-variables-path',
            str(manuscript_dir.joinpath('content/template-variables.json')),
        ])
    process = subprocess.run(
        args,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    print(process.args)
    print(process.stderr.decode())
    assert process.returncode == 0



import pytest

from manubot.cite.doi import (
    expand_short_doi,
)


def test_expand_short_doi():
    doi = expand_short_doi('10/b6vnmd')
    assert doi == "10.1016/s0933-3657(96)00367-3"


def test_expand_short_doi_invalid():
    with pytest.raises(ValueError, match='Handle not found. Double check short_doi'):
        expand_short_doi('10/b6vnmdxxxxxx')


def test_expand_short_doi_not_short():
    with pytest.raises(ValueError, match='shortDOIs start with `10/`'):
        expand_short_doi('10.1016/S0933-3657(96)00367-3')



import pathlib
import re

import setuptools

directory = pathlib.Path(__file__).parent.resolve()

# version
init_path = directory.joinpath('manubot', '__init__.py')
text = init_path.read_text()
pattern = re.compile(r"^__version__ = ['\"]([^'\"]*)['\"]", re.MULTILINE)
version = pattern.search(text).group(1)

# long_description
readme_path = directory.joinpath('README.md')
long_description = readme_path.read_text()

setuptools.setup(
    # Package details
    name='manubot',
    version=version,
    url='https://github.com/manubot/manubot',
    description='Manuscript bot for automated scientific publishing',
    long_description_content_type='text/markdown',
    long_description=long_description,
    license='BSD 3-Clause',

    # Author details
    author='Daniel Himmelstein',
    author_email='daniel.himmelstein@gmail.com',

    # Package topics
    keywords='manuscript markdown publishing references citations',
    classifiers=[
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
    ],

    packages=setuptools.find_packages(
        exclude=['tests']
    ),

    # Specify python version
    python_requires='>=3.6',

    # Run-time dependencies
    install_requires=[
        'errorhandler',
        'isbnlib',
        'jinja2',
        'jsonref',
        'jsonschema',
        'pandas',
        'pybase62',
        'pyyaml',
        'ratelimiter',
        'requests-cache',
        'requests',
    ],

    # Additional groups of dependencies
    extras_require={
    },

    # Create command line script
    entry_points={
        'console_scripts': [
            'manubot = manubot.command:main',
        ],
    },

    # Include package data files from MANIFEST.in
    include_package_data=True,
)



import importlib
import platform
import sys

# Email address that forwards to Manubot maintainers
contact_email = 'contact@manubot.org'


def import_function(name):
    """
    Import a function in a module specified by name. For example, if name were
    'manubot.cite.cite_command.cli_cite', the cli_cite function would be
    returned as an object. See https://stackoverflow.com/a/8790232/4651668.
    """
    module_name, function_name = name.rsplit('.', 1)
    module = importlib.import_module(module_name)
    return getattr(module, function_name)


def get_manubot_user_agent():
    """
    Return a User-Agent string for web request headers to help services
    identify requests as coming from Manubot.
    """
    try:
        from manubot import __version__ as manubot_version
    except ImportError:
        manubot_version = ''
    return (
        f'manubot/{manubot_version} '
        f'({platform.system()}; Python/{sys.version_info.major}.{sys.version_info.minor}) '
        f'<{contact_email}>'
    )



import functools
import logging
import re

from manubot.util import import_function


citeproc_retrievers = {
    'doi': 'manubot.cite.doi.get_doi_citeproc',
    'pmid': 'manubot.cite.pubmed.get_pubmed_citeproc',
    'pmcid': 'manubot.cite.pubmed.get_pmc_citeproc',
    'arxiv': 'manubot.cite.arxiv.get_arxiv_citeproc',
    'isbn': 'manubot.cite.isbn.get_isbn_citeproc',
    'wikidata': 'manubot.cite.wikidata.get_wikidata_citeproc',
    'url': 'manubot.cite.url.get_url_citeproc',
}

"""
Regex to extract citations.

Same rules as pandoc, except more permissive in the following ways:

1. the final character can be a slash because many URLs end in a slash.
2. underscores are allowed in internal characters because URLs, DOIs, and
   citation tags often contain underscores.

If a citation string does not match this regex, it can be substituted for a
tag that does, as defined in citation-tags.tsv.

https://github.com/greenelab/manubot-rootstock/issues/2#issuecomment-312153192

Prototyped at https://regex101.com/r/s3Asz3/2
"""
citation_pattern = re.compile(
    r'(?<!\w)@[a-zA-Z0-9][\w:.#$%&\-+?<>~/]*[a-zA-Z0-9/]')


@functools.lru_cache(maxsize=5_000)
def standardize_citation(citation):
    """
    Standardize citation identifiers based on their source
    """
    source, identifier = citation.split(':', 1)

    if source == 'doi':
        if identifier.startswith('10/'):
            from manubot.cite.doi import expand_short_doi
            try:
                identifier = expand_short_doi(identifier)
            except Exception as error:
                # If DOI shortening fails, return the unshortened DOI.
                # DOI metadata lookup will eventually fail somewhere with
                # appropriate error handling, as opposed to here.
                logging.error(
                    f'Error in expand_short_doi for {identifier} '
                    f'due to a {error.__class__.__name__}:\n{error}'
                )
                logging.info(error, exc_info=True)
        identifier = identifier.lower()

    if source == 'isbn':
        from isbnlib import to_isbn13
        identifier = to_isbn13(identifier)

    return f'{source}:{identifier}'


regexes = {
    'pmid': re.compile(r'[1-9][0-9]{0,7}'),
    'pmcid': re.compile(r'PMC[0-9]+'),
    'doi': re.compile(r'10\.[0-9]{4,9}/\S+'),
    'shortdoi': re.compile(r'10/[a-zA-Z0-9]+'),
    'wikidata': re.compile(r'Q[0-9]+'),
}


def inspect_citation_identifier(citation):
    """
    Check citation identifiers adhere to expected formats. If an issue is
    detected a string describing the issue is returned. Otherwise returns None.
    """
    source, identifier = citation.split(':', 1)

    if source == 'pmid':
        # https://www.nlm.nih.gov/bsd/mms/medlineelements.html#pmid
        if identifier.startswith('PMC'):
            return (
                'PubMed Identifiers should start with digits rather than PMC. '
                f'Should {citation} switch the citation source to `pmcid`?'
            )
        elif not regexes['pmid'].fullmatch(identifier):
            return 'PubMed Identifiers should be 1-8 digits with no leading zeros.'

    if source == 'pmcid':
        # https://www.nlm.nih.gov/bsd/mms/medlineelements.html#pmc
        if not identifier.startswith('PMC'):
            return 'PubMed Central Identifiers must start with `PMC`.'
        elif not regexes['pmcid'].fullmatch(identifier):
            return (
                'Identifier does not conform to the PMCID regex. '
                'Double check the PMCID.'
            )

    if source == 'doi':
        if identifier.startswith('10.'):
            # https://www.crossref.org/blog/dois-and-matching-regular-expressions/
            if not regexes['doi'].fullmatch(identifier):
                return (
                    'Identifier does not conform to the DOI regex. '
                    'Double check the DOI.'
                )
        elif identifier.startswith('10/'):
            # shortDOI, see http://shortdoi.org
            if not regexes['shortdoi'].fullmatch(identifier):
                return (
                    'Identifier does not conform to the shortDOI regex. '
                    'Double check the shortDOI.'
                )
        else:
            return (
                'DOIs must start with `10.` (or `10/` for shortDOIs).'
            )

    if source == 'isbn':
        import isbnlib
        fail = isbnlib.notisbn(identifier, level='strict')
        if fail:
            return (
                f'identifier violates the ISBN syntax according to isbnlib v{isbnlib.__version__}'
            )

    if source == 'wikidata':
        # https://www.wikidata.org/wiki/Wikidata:Identifiers
        if not identifier.startswith('Q'):
            return (
                'Wikidata item IDs must start with `Q`.'
            )
        elif not regexes['wikidata'].fullmatch(identifier):
            return (
                'Identifier does not conform to the Wikidata regex. '
                'Double check the entity ID.'
            )

    return None


def is_valid_citation_string(
        string, allow_tag=False, allow_raw=False, allow_pandoc_xnos=False):
    """
    Return True if string is a properly formatted citation. Return False if
    string is not a citation or is an invalid citation.

    In the case string is an invalid citation, an error is logged. This
    function does not catch all invalid citations, but instead performs cursory
    checks, such as citations adhere to the expected formats. No calls to
    external resources are used by these checks, so they will not detect
    citations to non-existent identifiers unless those identifiers violate
    their source's syntax.

    allow_tag=False, allow_raw=False, and allow_pandoc_xnos=False enable
    allowing citation sources that are valid for Manubot manuscripts, but
    likely not elsewhere. allow_tag=True enables citations tags (e.g.
    @tag:citation-tag). allow_raw=True enables raw citations (e.g.
    @raw:manual-reference). allow_pandoc_xnos=True still returns False for
    pandoc-xnos references (e.g. @fig:figure-id), but does not log an error.
    With the default of False for these arguments, valid sources are restricted
    to those for which manubot can retrieve metadata based only on the
    standalone citation.
    """
    if not string.startswith('@'):
        logging.error(f'Invalid citation: {string}\ndoes not start with "@"')
        return False
    citation = string[1:]
    try:
        source, identifier = citation.split(':', 1)
    except ValueError as e:
        logging.error(
            f'Citation not splittable via a single colon: {string}. '
            'Citation strings must be in the format of `@source:identifier`.'
        )
        return False

    if not source or not identifier:
        msg = f'Invalid citation: {string}\nblank source or identifier'
        logging.error(msg)
        return False

    if allow_pandoc_xnos:
        # Exempted non-citation sources used for pandoc-fignos,
        # pandoc-tablenos, and pandoc-eqnos
        pandoc_xnos_keys = {'fig', 'tbl', 'eq'}
        if source in pandoc_xnos_keys:
            return False
        if source.lower() in pandoc_xnos_keys:
            logging.error(
                f'pandoc-xnos reference types should be all lowercase.\n'
                f'Should {string} use "{source.lower()}" rather than "{source}"?'
            )
            return False

    # Check supported source type
    sources = set(citeproc_retrievers)
    if allow_raw:
        sources.add('raw')
    if allow_tag:
        sources.add('tag')
    if source not in sources:
        if source.lower() in sources:
            logging.error(
                f'Citation sources should be all lowercase.\n'
                f'Should {string} use "{source.lower()}" rather than "{source}"?'
            )
        else:
            logging.error(
                f'Invalid citation: {string}\n'
                f'Source "{source}" is not valid.\n'
                f'Valid citation sources are {{{", ".join(sorted(sources))}}}'
            )
        return False

    inspection = inspect_citation_identifier(citation)
    if inspection:
        logging.error(f'invalid {source} citation: {string}\n{inspection}')
        return False

    return True


def get_citation_id(standard_citation):
    """
    Get the citation_id for a standard_citation.
    """
    import hashlib
    import base62
    assert '@' not in standard_citation
    as_bytes = standard_citation.encode()
    blake_hash = hashlib.blake2b(as_bytes, digest_size=6)
    digest = blake_hash.digest()
    citation_id = base62.encodebytes(digest)
    return citation_id


def citation_to_citeproc(citation, prune=True):
    """
    Return a dictionary with citation metadata
    """
    assert citation == standardize_citation(citation)
    source, identifier = citation.split(':', 1)

    if source in citeproc_retrievers:
        citeproc_retriever = import_function(citeproc_retrievers[source])
        citeproc = citeproc_retriever(identifier)
    else:
        msg = f'Unsupported citation source {source} in {citation}'
        raise ValueError(msg)

    citation_id = get_citation_id(citation)
    from manubot.cite.citeproc import citeproc_passthrough
    citeproc = citeproc_passthrough(citeproc, set_id=citation_id, prune=prune)

    return citeproc



import pytest

from manubot.cite.zotero import (
    web_query,
    search_query,
    export_as_csl
)


def test_web_query():
    """
    The translation-server web endpoint can be tested via curl:
    ```
    curl \
      --header "Content-Type: text/plain" \
      --data 'https://bigthink.com/neurobonkers/a-pirate-bay-for-science' \
      'https://translate.manubot.org/web'
    ```
    An outdated installation of translation-server caused the web query for this
    URL to be extraordinarily slow but has now been fixed. See
    https://github.com/zotero/translation-server/issues/63
    """
    url = 'https://bigthink.com/neurobonkers/a-pirate-bay-for-science'
    zotero_data = web_query(url)
    assert isinstance(zotero_data, list)
    assert len(zotero_data) == 1
    assert zotero_data[0]['title'] == "Meet the Robin Hood of Science"


def test_export_as_csl():
    """
    CSL export can be tested via curl:
    ```
    curl \
      --header "Content-Type: application/json" \
      --data '[{"key": "IN22XN53", "itemType": "webpage", "date": "2016-02-09T20:12:00"}]' \
      'https://translate.manubot.org/export?format=csljson'
    ```
    """
    zotero_data = [
        {
            "key": "IN22XN53",
            "version": 0,
            "itemType": "webpage",
            "creators": [],
            "tags": [],
            "title": "Meet the Robin Hood of Science",
            "websiteTitle": "Big Think",
            "date": "2016-02-09T20:12:00",
            "url": "https://bigthink.com/neurobonkers/a-pirate-bay-for-science",
            "abstractNote": "How one researcher created a pirate bay for science more powerful than even libraries at top universities.",
            "language": "en",
            "accessDate": "2018-12-06T20:10:14Z",
        }
    ]
    csl_item = export_as_csl(zotero_data)[0]
    assert csl_item['title'] == "Meet the Robin Hood of Science"
    assert csl_item['container-title'] == 'Big Think'


def test_web_query_returns_single_result_legacy_manubot_url():
    """
    Check that single=1 is specified for web queries. Without this, Zotero
    can prefer translators that return multiple choices. This occurs with legacy
    Manubot mansucripts, which get assigned the DOI translator as top priority.
    https://github.com/zotero/translation-server/issues/65
    ```
    curl \
      --header "Content-Type: text/plain" \
      --data 'https://greenelab.github.io/scihub-manuscript/v/cfe599e25405d38092bf972b6ea1c9e0dcf3deb9/' \
      'https://translate.manubot.org/web?single=1'
    ```
    """
    url = 'https://greenelab.github.io/scihub-manuscript/v/cfe599e25405d38092bf972b6ea1c9e0dcf3deb9/'
    zotero_metadata = web_query(url)
    assert isinstance(zotero_metadata, list)
    assert len(zotero_metadata) == 1
    zotero_metadata, = zotero_metadata
    assert zotero_metadata['title'] == 'Sci-Hub provides access to nearly all scholarly literature'


def test_web_query_returns_single_result_pubmed_url():
    """
    See test_web_query_returns_single_result_legacy_manubot_url docstring.
    ```
    curl \
      --header "Content-Type: text/plain" \
      --data 'https://www.ncbi.nlm.nih.gov/pubmed/?term=sci-hub%5Btitle%5D' \
      'https://translate.manubot.org/web?single=1'
    ```
    """
    url = 'https://www.ncbi.nlm.nih.gov/pubmed/?term=sci-hub%5Btitle%5D'
    zotero_metadata = web_query(url)
    assert isinstance(zotero_metadata, list)
    assert len(zotero_metadata) == 1
    zotero_metadata, = zotero_metadata
    assert zotero_metadata['title'] == 'sci-hub[title] - PubMed - NCBI'


def test_search_query_isbn():
    """
    The translation-server search endpoint can be tested via curl:
    ```
    curl \
      --header "Content-Type: text/plain" \
      --data 'isbn:9781339919881' \
      'https://translate.manubot.org/search'
    ```
    """
    identifier = 'isbn:9781339919881'
    zotero_data = search_query(identifier)
    assert zotero_data[0]['title'].startswith('The hetnet awakens')


def test_search_query_arxiv():
    """
    Test citing https://arxiv.org/abs/1604.05363v1
    The translation-server search endpoint can be tested via curl:
    ```
    curl --verbose \
      --header "Content-Type: text/plain" \
      --data 'arxiv:1604.05363v1' \
      'https://translate.manubot.org/search'
    ```
    """
    identifier = 'arxiv:1604.05363v1'
    zotero_data = search_query(identifier)
    assert zotero_data[0]['title'] == 'Comparing Published Scientific Journal Articles to Their Pre-print Versions'
    assert zotero_data[0]['creators'][-1]['firstName'] == 'Todd'
    assert zotero_data[0]['date'] == '2016-04-18'


@pytest.mark.parametrize('identifier', [
    '30571677',  # https://www.ncbi.nlm.nih.gov/pubmed/30571677
    'doi:10.1371/journal.pcbi.1006561',  # https://doi.org/10.1371/journal.pcbi.1006561
])
def test_search_query(identifier):
    """
    The translation-server search endpoint can be tested via curl:
    ```
    curl --verbose \
      --header "Content-Type: text/plain" \
      --data '30571677' \
      'https://translate.manubot.org/search'
    ```
    translation-server does not support PMIDs with a `pmid:` prefix.
    https://github.com/zotero/translation-server/issues/71
    """
    zotero_data = search_query(identifier)
    assert zotero_data[0]['title'].startswith('Ten simple rules for documenting scientific software')
    assert zotero_data[0]['creators'][0]['lastName'] == 'Lee'



import json
import logging
import pathlib
import shutil
import subprocess
import sys

from manubot.cite import (
    citation_to_citeproc,
    standardize_citation,
)
from manubot.cite.util import is_valid_citation_string

# For manubot cite, infer --format from --output filename extensions
extension_to_format = {
    '.txt': 'plain',
    '.md': 'markdown',
    '.docx': 'docx',
    '.html': 'html',
    '.xml': 'jats',
}


def call_pandoc(metadata, path, format='plain'):
    """
    path is the path to write to.
    """
    info = _get_pandoc_info()
    _check_pandoc_version(info, metadata, format)
    metadata_block = '---\n{yaml}\n...\n'.format(
        yaml=json.dumps(metadata, ensure_ascii=False, indent=2)
    )
    args = [
        'pandoc',
        '--filter', 'pandoc-citeproc',
        '--output', str(path) if path else '-',
    ]
    if format == 'markdown':
        args.extend(['--to', 'markdown_strict', '--wrap', 'none'])
    elif format == 'jats':
        args.extend(['--to', 'jats', '--standalone'])
    elif format == 'docx':
        args.extend(['--to', 'docx'])
    elif format == 'html':
        args.extend(['--to', 'html'])
    elif format == 'plain':
        args.extend(['--to', 'plain', '--wrap', 'none'])
        if info['pandoc version'] >= (2,):
            # Do not use ALL_CAPS for bold & underscores for italics
            # https://github.com/jgm/pandoc/issues/4834#issuecomment-412972008
            filter_path = pathlib.Path(__file__).joinpath('..', 'plain-pandoc-filter.lua').resolve()
            assert filter_path.exists()
            args.extend(['--lua-filter', str(filter_path)])
    logging.info('call_pandoc subprocess args:\n' + ' '.join(args))
    process = subprocess.run(
        args=args,
        input=metadata_block.encode(),
        stdout=subprocess.PIPE if path else sys.stdout,
        stderr=sys.stderr,
    )
    process.check_returncode()


def cli_cite(args):
    """
    Main function for the manubot cite command-line interface.

    Does not allow user to directly specify Pandoc's --to argument, due to
    inconsistent citaiton rendering by output format. See
    https://github.com/jgm/pandoc/issues/4834
    """
    # generate CSL JSON data
    csl_list = list()
    for citation in args.citations:
        try:
            if not is_valid_citation_string(f'@{citation}'):
                continue
            citation = standardize_citation(citation)
            csl_item = citation_to_citeproc(citation, prune=args.prune_csl)
            csl_list.append(csl_item)
        except Exception as error:
            logging.error(
                f'citation_to_citeproc for {citation} failed '
                f'due to a {error.__class__.__name__}:\n{error}'
            )
            logging.info(error, exc_info=True)

    # output CSL JSON data, if --render is False
    if not args.render:
        write_file = args.output.open('w') if args.output else sys.stdout
        with write_file:
            json.dump(csl_list, write_file, ensure_ascii=False, indent=2)
            write_file.write('\n')
        return

    # use Pandoc to render citations
    if not args.format and args.output:
        vars(args)['format'] = extension_to_format.get(args.output.suffix)
    if not args.format:
        vars(args)['format'] = 'plain'
    pandoc_metadata = {
        'nocite': '@*',
        'csl': args.csl,
        'references': csl_list,
    }
    call_pandoc(
        metadata=pandoc_metadata,
        path=args.output,
        format=args.format,
    )


def _get_pandoc_info():
    """
    Return path and version information for the system's pandoc and
    pandoc-citeproc commands. If not available, exit program.
    """
    stats = dict()
    for command in 'pandoc', 'pandoc-citeproc':
        path = shutil.which(command)
        if not path:
            logging.critical(
                f'"{command}" not found on system. '
                f'Check that Pandoc is installed.'
            )
            raise SystemExit(1)
        version = subprocess.check_output(
            args=[command, '--version'],
            universal_newlines=True,
        )
        logging.debug(version)
        version, *discard = version.splitlines()
        discard, version = version.strip().split()
        version = tuple(map(int, version.split('.')))
        stats[f'{command} version'] = version
        stats[f'{command} path'] = path
    logging.info('\n'.join(f'{k}: {v}' for k, v in stats.items()))
    return stats


def _check_pandoc_version(info, metadata, format):
    """
    Given info from _get_pandoc_info, check that Pandoc's version is sufficient
    to perform the citation rendering command specified by metadata and format.
    Please add additional minimum version information to this function, as its
    discovered.
    """
    issues = list()
    if format == 'jats' and info['pandoc version'] < (2,):
        issues.append('--jats requires pandoc >= v2.0.')
    # --csl=URL did not work in https://travis-ci.org/greenelab/manubot/builds/417314743#L796, but exact version where this fails unknown
    # if metadata.get('csl', '').startswith('http') and pandoc_version < (2,):
    #     issues.append('--csl=URL requires pandoc >= v2.0.')
    issues = '\n'.join(issues)
    if issues:
        logging.critical(f'issues with pandoc version detected:\n{issues}')
